# -*- coding: utf-8 -*
from bw2data.utils import MAX_INT_32
import numpy as np
import itertools


def indexer(array_from, array_to, mapping):
    """
Chosen algorithm is faster than using rankdata, or simple iteration.
Modifies numpy arrays in place.

map(dict.get, data) seems faster, but doesn't work with missing values.

Setup code to try new approaches::

    from numpy import *
    from time import time
    a = random.random_integers(5000, size=100000)
    indices = sorted(list(set([int(x) for x in a])))
    mapping = dict(zip(indices, range(len(indices))))
    b = zeros(a.shape[0])

    """
    array_to[:] = np.array([mapping.get(x, MAX_INT_32) for x in array_from])


def dicter(array):
    """
Create dictionary from the sorted, unique values in a numpy array

Already more than fast enough. Setup code for other approaches::

        %%timeit import numpy as np; array = np.random.randint(0, 5000, size=50000)
        dict(zip(np.sort(np.unique(array)), itertools.count()))

    """
    return dict(zip(np.sort(np.unique(array)), itertools.count()))



# -*- coding: utf-8 -*
from __future__ import division
from .lca import LCA
from bw2data.proxies import CompressedSparseMatrixProxy
from scipy import sparse
import numpy as np


class SimpleRegionalizedLCA(LCA):
    """A semi-regionalized LCA calculation where the inventory and impact assessment spatial scales match."""

    def build_characterization_matrix(self, vector=None):
        """Build a regionalized characterization matrix.

        Throughout this calculation, locations refer to integer indices retrieved from the geomapper, not the string codes themselves. We operate directly on the paramter arrays, as this is much more efficient.

        We do this by first retrieving the regionalized characterization factors for each location where characterization factors are avaiable. The intermediate data structure ``regionalized_dict`` has the following structure:

        .. code-block:: python

            {location_id: (biosphere matrix row number, cf value)}

        We then use the ``np.unique`` function to retrieve all technosphere processes, and an index into ``self.tech_params`` for each of them. We can then use this index to get a location for each technosphere process.

        The characterization matrix has dimensions (number of biosphere flows, number of technosphere flows). For each column, we lookup the location code, and then retrieve the cf amounts and row indices from the ``regionalized_dict``. We can then build the ``characterization_matrix``.

        .. note:: There is a lot of duplicate data in ``characterization_matrix``, as characterization factors are provided for each technosphere process, regardless of whether that technosphere location has been seen already.

        """
        vector = self.cf_params['amount'] if vector is None else vector
        # count = len(self.biosphere_dict)
        regionalized_dict = {}
        for index in vector.shape[0]:
            regionalized_dict.setdefault(
                int(self.cf_params["geo"][index]), []).append((
                    int(self.cf_params["index"][index]), vector[index]))
        for key, data in regionalized_dict.iteritems():
            regionalized_dict[key] = (
                np.array([x[0] for x in regionalized_dict]),
                np.array([x[1] for x in regionalized_dict]))
        # Get location codes for technosphere processes
        tech_columns, tech_param_indices = np.unique(
            self.tech_params["col"], return_index=True)
        tech_geo_indices = self.tech_params["geo"][tech_param_indices]
        cfs, rows, cols = [], [], []
        for i, col_index in enumerate(tech_columns):
            geo = int(tech_geo_indices[i])
            these_cfs = regionalized_dict[geo][1]
            cfs.append(these_cfs)
            cols.append(np.ones(these_cfs.shape[0]) * col_index)
            rows.append(regionalized_dict[geo][0])
        self.characterization_matrix = sparse.coo_matrix(
            (np.hstack(cfs), (np.hstack(rows), np.hstack(cols))),
            (len(self.biosphere_dict), len(self.technosphere_dict))
        ).tocsr()

    def lcia_calculation(self):
        """Multiply the characterization matrix by the life cycle inventory. Uses ``.multiply`` for point-wise multiplication, as both matrices have the same dimensions."""
        self.characterized_inventory = CompressedSparseMatrixProxy(
            self.characterization_matrix.multiply(self.inventory.data),
            self.biosphere_dict, self.technosphere_dict)



# -*- coding: utf-8 -*
from .lca import LCA
from .simple_regionalized import SimpleRegionalizedLCA
from .monte_carlo import MonteCarloLCA, ParallelMonteCarlo, MultiMonteCarlo
from .graph_traversal import GraphTraversal, edge_cutter, node_pruner, \
    d3_fd_graph_formatter
from .matrices import MatrixBuilder, TechnosphereBiosphereMatrixBuilder

__version__ = (0, 9, 3)



from .. import LCA
from bw2data import *
from bw2data.tests import BW2DataTest
import numpy as np


class LCACalculationTestCase(BW2DataTest):
    def add_basic_biosphere(self):
        biosphere = Database("biosphere")
        biosphere.register("Made for tests", [], 1)
        biosphere.write({
            ("biosphere", 1): {
                'categories': ['things'],
                'code': 1,
                'exchanges': [],
                'name': 'an emission',
                'type': 'emission',
                'unit': 'kg'
                }})
        biosphere.process()

    def test_basic(self):
        test_data = {
            ("t", 1): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', 2),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', 1),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", 2): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register("Made for tests", ["biosphere"], 2)
        test_db.write(test_data)
        test_db.process()
        lca = LCA({("t", 1): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.technosphere_dict[mapping[("t", 1)]]] = 1
        answer[lca.technosphere_dict[mapping[("t", 2)]]] = 0.5
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_production_values(self):
        test_data = {
            ("t", 1): {
                'exchanges': [{
                    'amount': 2,
                    'input': ('t', 1),
                    'type': 'production',
                    'uncertainty type': 0},
                    {'amount': 0.5,
                    'input': ('t', 2),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', 1),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", 2): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register("Made for tests", ["biosphere"], 2)
        test_db.write(test_data)
        test_db.process()
        lca = LCA({("t", 1): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.technosphere_dict[mapping[("t", 1)]]] = 0.5
        answer[lca.technosphere_dict[mapping[("t", 2)]]] = 0.25
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_substitution(self):
        test_data = {
            ("t", 1): {
                'exchanges': [{
                    'amount': -1,  # substitution
                    'input': ('t', 2),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', 1),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", 2): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register("Made for tests", ["biosphere"], 2)
        test_db.write(test_data)
        test_db.process()
        lca = LCA({("t", 1): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.technosphere_dict[mapping[("t", 1)]]] = 1
        answer[lca.technosphere_dict[mapping[("t", 2)]]] = -1
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_circular_chains(self):
        test_data = {
            ("t", 1): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', 2),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', 1),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", 2): {
                'exchanges': [{
                    'amount': 0.1,
                    'input': ('t', 1),
                    'type': 'technosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register("Made for tests", ["biosphere"], 2)
        test_db.write(test_data)
        test_db.process()
        lca = LCA({("t", 1): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.technosphere_dict[mapping[("t", 1)]]] = 20 / 19.
        answer[lca.technosphere_dict[mapping[("t", 2)]]] = 10 / 19.
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_dependent_databases(self):
        pass

    def test_demand_type(self):
        with self.assertRaises(ValueError):
            LCA(("foo", 1))
        with self.assertRaises(ValueError):
            LCA("foo")
        with self.assertRaises(ValueError):
            LCA([{"foo": 1}])

    def test_decomposed_uses_solver(self):
        test_data = {
            ("t", 1): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', 2),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', 1),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", 2): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register("Made for tests", ["biosphere"], 2)
        test_db.write(test_data)
        test_db.process()
        lca = LCA({("t", 1): 1})
        lca.lci(factorize=True)
        # Indirect test because no easy way to test a function is called
        lca.technosphere_matrix = None
        self.assertEqual(float(lca.solve_linear_system().sum()), 1.5)



# -*- coding: utf-8 -*
from __future__ import division
from . import LCA
from brightway2 import mapping, Database
from heapq import heappush, heappop
import numpy as np


class GraphTraversal(object):
    """Master class for graph traversal."""
    def calculate(self, demand, method, cutoff=0.005, max_calc=1e5):
        counter = 0

        lca, supply, score = self.build_lca(demand, method)
        if score == 0:
            raise ValueError("Zero total LCA score makes traversal impossible")

        # Create matrix of LCIA CFs times biosphere flows, as these don't
        # change. This is also the unit score of each activity.
        characterized_biosphere = np.array((
            lca.characterization_matrix.data *
            lca.biosphere_matrix.data).sum(axis=0)).ravel()

        heap, nodes, edges = self.initialize_heap(
            demand, lca, supply, characterized_biosphere)
        nodes, edges, counter = self.traverse(
            heap, nodes, edges, counter, max_calc, cutoff, score, supply,
            characterized_biosphere, lca)
        nodes = self.add_metadata(nodes, lca)

        return {
            'nodes': nodes,
            'edges': edges,
            'lca': lca,
            'counter': counter,
        }

    def initialize_heap(self, demand, lca, supply, characterized_biosphere):
        heap, nodes, edges = [], {}, []
        for activity, value in demand.iteritems():
            index = lca.technosphere_dict[mapping[activity]]
            heappush(heap, (1, index))
            nodes[index] = {
                "amount": supply[index],
                "cum": self.cumulative_score(
                    index, supply, characterized_biosphere, lca),
                "ind": self.unit_score(index, supply, characterized_biosphere)
            }
            # -1 is a special index for total demand, which can be
            # composite. Initial edges are inputs to the
            # functional unit.
            edges.append({
                "to": -1,
                "from": index,
                "amount": value,
                "impact": lca.score,
            })
        return heap, nodes, edges

    def build_lca(self, demand, method):
        lca = LCA(demand, method)
        lca.lci()
        lca.lcia()
        lca.decompose_technosphere()
        return lca, lca.solve_linear_system(), lca.score

    def cumulative_score(self, index, supply, characterized_biosphere, lca):
        demand = np.zeros((supply.shape[0],))
        demand[index] = supply[index]
        return float((characterized_biosphere * lca.solver(demand)).sum())

    def unit_score(self, index, supply, characterized_biosphere):
        return float(characterized_biosphere[index] * supply[index])

    def traverse(self, heap, nodes, edges, counter, max_calc, cutoff,
                 total_score, supply, characterized_biosphere, lca):
        """
Build a directed graph of the supply chain.

Use a heap queue to store a sorted list of processes that need to be examined,
and traverse the graph using an "importance-first" search.
        """
        while heap and counter < max_calc:
            parent_score_inverted, parent_index = heappop(heap)
            # parent_score = 1 / parent_score_inverted
            col = lca.technosphere_matrix.data[:, parent_index].tocoo()
            # Multiply by -1 because technosphere values are negative
            # (consumption of inputs)
            children = [(col.row[i], -1 * col.data[i]) for i in xrange(
                col.row.shape[0])]
            for activity, amount in children:
                # Skip values on technosphere diagonal or coproducts
                if activity == parent_index or amount <= 0:
                    continue
                counter += 1
                cumulative_score = self.cumulative_score(
                    activity, supply, characterized_biosphere, lca)
                if abs(cumulative_score) < abs(total_score * cutoff):
                    continue
                # Edge format is (to, from, mass amount, cumulative impact)
                edges.append({
                    "to": parent_index,
                    "from": activity,
                    # The cumulative impact directly due to this link (weight)
                    # Amount of this link * amount of parent demanding link
                    "amount": amount * nodes[parent_index]["amount"],
                    # Amount of this input
                    "impact": amount * nodes[parent_index]["amount"]
                    # times impact per unit of this input
                    * cumulative_score / supply[activity]
                })
                # Want multiple incoming edges, but don't add existing node
                if activity in nodes:
                    continue
                nodes[activity] = {
                    # Total amount of this flow supplied
                    "amount": supply[activity],
                    # Cumulative score from all flows of this activity
                    "cum": cumulative_score,
                    # Individual score attributable to environmental flows
                    # coming directory from or to this activity
                    "ind": self.unit_score(activity, supply,
                                           characterized_biosphere)
                }
                heappush(heap, (1 / cumulative_score, activity))

        return nodes, edges, counter

    def add_metadata(self, nodes, lca):
        rm = dict([(v, k) for k, v in mapping.data.iteritems()])
        rt = dict([(v, k) for k, v in lca.technosphere_dict.iteritems()])
        lookup = dict([(index, self.get_code(index, rm, rt)) for index in nodes if index != -1])
        new_nodes = [(-1, {
            "code": "fu",
            "cum": lca.score,
            "ind": 1e-6 * lca.score,
            "amount": 1,
            "name": "Functional unit",
            "cat": "Functional unit"
        })]
        for key, value in nodes.iteritems():
            if key == -1:
                continue
            code = lookup[key]
            db_data = Database(code[0]).load()
            value.update({
                "code": code,
                "name": db_data[code]["name"],
                "cat": db_data[code]["categories"][0],
            })
            new_nodes.append((key, value))
        return dict(new_nodes)

    def get_code(self, index, rev_mapping, rev_tech):
        return rev_mapping[rev_tech[index]]


def edge_cutter(nodes, edges, total, limit=0.0025):
    """The default graph traversal includes links which might be of small magnitude. This function cuts links that have small cumulative impact."""
    to_delete = []
    for i, e in enumerate(edges):
        if e["impact"] < (total * limit):
            to_delete.append(i)
        else:
            continue
            # print e
            # print nodes[e[1]]
            # print (e[2] / nodes[e[1]]["amount"] * nodes[e[1]]["cum"]) / (total * limit)
    return [e for i, e in enumerate(edges) if i not in to_delete]


def node_pruner(nodes, edges):
    """Remove nodes which have no links remaining after edge cutting."""
    good_nodes = set([e["from"] for e in edges]).union(
        set([e["to"] for e in edges]))
    return dict([(k, v) for k, v in nodes.iteritems() if k in good_nodes])


def extract_edges(arr, mapping, ignore):
    edges = []
    for i in range(arr.shape[0]):
        if mapping[i] in ignore:
            continue
        for j in range(arr.shape[1]):
            if mapping[j] in ignore or i == j or arr[i, j] == 0:
                continue
            edges.append((mapping[j], mapping[i], float(arr[i, j])))
    return edges


def rationalize_supply_chain(nodes, edges, total, limit=0.005):
    """
This class takes nodes and edges, and removes nodes to edges with low individual scores and reroutes the edges.
    """
    nodes_to_delete = [key for key, value in nodes.iteritems() if
                       value["ind"] < (total * limit) and key != -1]
    size = len(nodes)
    arr = np.zeros((size, size), dtype=np.float32)
    arr_map = dict([(key, index) for index, key in enumerate(sorted(nodes.keys()))])
    rev_map = dict([(v, k) for k, v in arr_map.iteritems()])
    for outp, inp, amount in edges:
        arr[arr_map[inp], arr_map[outp]] = amount
    for node in nodes_to_delete:
        index = arr_map[node]
        increment = (arr[:, index].reshape((-1, 1)) * arr[index, :].reshape((1, -1)))
        arr += increment
    new_edges = []
    new_edges = extract_edges(arr, rev_map, nodes_to_delete)
    new_nodes = dict([(k, v) for k, v in nodes.iteritems() if k not in nodes_to_delete])
    return new_nodes, new_edges


def d3_fd_graph_formatter(nodes, edges, total):
        # Sort node ids by highest cumulative score first
        node_ids = [x[1] for x in sorted(
            [(v["cum"], k) for k, v in nodes.iteritems()])]
        new_nodes = [nodes[i] for i in node_ids]
        lookup = dict([(key, index) for index, key in enumerate(node_ids)])
        new_edges = [{
            "source": lookup[e["to"]],
            "target": lookup[e["from"]],
            "amount": e["impact"]
        } for e in edges]
        return {"edges": new_edges, "nodes": new_nodes,
                "total": total}



from setuptools import setup

setup(
    name='bw2calc',
    version="0.9.3",
    packages=["bw2calc", "bw2calc.tests"],
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license=open('LICENSE.txt').read(),
    url="https://bitbucket.org/cmutel/brightway2-calc",
    install_requires=["brightway2"],
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)
